import re
from typing import List, Tuple

SINGLE_QUOTES_MAP = str.maketrans({"‘": "'", "’": "'", "‛": "'", "❛": "'", "❜": "'"})
DOUBLE_QUOTES_MAP = str.maketrans(
    {
        "“": '"',
        "”": '"',
        "„": '"',
        "❠": '"',
        "〝": '"',
        "〞": '"',
        "〟": '"',
        "＂": '"',
    }
)


def get_quotes(text: str) -> list[int]:
    """
    Count and return the total quotes and their indexes in the given text.

    Args:
        text (str): The input text.

    Returns:
        Tuple[int, List[int]]: A tuple containing the total quotes and their indexes.
    """
    return [i for i, char in enumerate(text) if char == '"']


def clear_subquotes(text: str) -> str:
    """
    Remove sub-quotes and return the modified text.

    Args:
        text (str): The input text.

    Returns:
        str: The modified text without sub-quotes.
    """
    return re.sub("''", "'", text.translate(SINGLE_QUOTES_MAP))


def set_simple_quotes(text: str, simplify_everything: bool = False) -> str:
    """
    Replace special quotation marks with simple double quotes in the given text.

    Args:
        text (str): The input text.

    Returns:
        str: The text with special quotation marks replaced by simple double quotes.
    """
    res = re.sub('""', '"', text.translate(DOUBLE_QUOTES_MAP))
    if simplify_everything:
        return clear_subquotes(res)
    return res


def update_text(positions: List[int], text: str, q1: str = "“", q2: str = "”") -> str:
    """
    Update the text by enclosing specified positions with given quotation marks.

    Args:
        positions (List[int]): A list of positions where quotation marks should be applied.
        text (str): The input text.
        q1 (str, optional): The opening quotation mark. Defaults to "“".
        q2 (str, optional): The closing quotation mark. Defaults to "”".

    Returns:
        str: The updated text with specified positions enclosed by the given quotation marks.
    """
    split_version = [positions[i : i + 2] for i in range(0, len(positions), 2)]
    for pair in split_version:
        if len(pair) == 2:
            text = (
                text[: pair[0]]
                + q1
                + text[pair[0] + 1 : pair[1]]
                + q2
                + text[pair[1] + 1 :]
            )
        elif len(pair) == 1:
            text = text[: pair[0]] + "" + text[pair[0] + 1 :]
    return text


def requoter(input_text: str, use_simple: bool = False) -> str:
    """
    Process the input text, fixing and changing the quotation style.

    Args:
        input_text (str): The text input that will be converted.
        use_simple (bool, optional): If `True`, It will return the text with the simply quotes instead of the special markings. Defaults to False.

    Returns:
        str: The processed text with fixed and changed quotation style.
    """
    input_text = set_simple_quotes(input_text, use_simple)
    if use_simple:
        return input_text

    positions = get_quotes(input_text)
    if not positions:
        return input_text.replace("'", "’")
    return update_text(positions, input_text).replace("'", "’")



import json
from typing import Optional, Union
from typing import List, Dict, Callable, Any, Optional

# To make this as optional, this was setup.
pydantic_available = False
try:
    from pydantic import BaseModel  # type: ignore

    pydantic_available = True
except ModuleNotFoundError:

    class BaseModel:
        pass


def safe_execute(func: Callable, *args, **kwargs) -> Dict[str, Any]:
    """Executes a function safely, returning errors as a dictionary."""
    try:
        result = func(*args, **kwargs)
        return {"success": True, "result": result}
    except Exception as e:
        return {"success": False, "error": str(e)}


# Validation models
class Action(BaseModel):
    item: str
    args: List[Any] = []
    kwargs: Dict[str, Any] = {}
    result_var: str
    post_processor: Optional[str] = None


class ReturnConfig(BaseModel):
    fn: Optional[str] = None
    args: List[str] = []
    result_var: str


class Command(BaseModel):
    actions: List[Action]
    returns: Optional[Dict[str, ReturnConfig]] = None


# Interpreter class
class CommandInterpreter:
    """Example:
        ```python
        from gr1336_toolbox import (
            compare_texts,
            text_entropy,
            normalize_text,
            Command,
            CommandInterpreter,
        )

        # function map
        function_map = {
            "compare_texts": compare_texts,
            "text_entropy": text_entropy,
            "normalize_text": normalize_text,
        }

    if __name__ == "__main__":
        example_input = {
            "actions": [
                {
                    "item": "normalize_text",
                    "args": ["This is a simple test."],
                    "kwargs": {},
                    "result_var": "normalized_text",
                    "post_processor": None,
                },
                {
                    "item": "text_entropy",
                    "args": [],
                    "kwargs": {"text": "This is a simple test"},
                    "result_var": "text_entropy_value",
                    "post_processor": None,
                },
                {
                    "item": "compare_texts",
                    "args": [],
                    "kwargs": {"text1": "hello world\\nhi there\\nyesterday", "text2": "hello there\\nhi there\\ntoday"},
                    "result_var": "text_comparison",
                    "post_processor": None,
                },
            ],
            "returns": {
                "post_process": {
                    "fn": "lambda x, y: f'Normalized: {x}, Entropy: {y}'",
                    "args": ["normalized_text", "text_entropy_value"],
                    "result_var": "summary",
                }
            },
        }

        interpreter = CommandInterpreter(Command, function_map)
        result_context = interpreter.execute(example_input)
        print(result_context)

        # Results printout:
        result_context = {
            "normalized_text": {"success": True, "result": "this is a simple test."},
            "text_entropy_value": {"success": True, "result": 3.08232813685843},
            "text_comparison": {
                "success": True,
                "result": ["- hello world", "+ hello there", "- yesterday", "+ today"],
            },
            "summary": {
                "success": True,
                "result": "Normalized: this is a simple test., Entropy: 3.08232813685843",
            },
        }
        ```
    """
    def __init__(self, validation_model: BaseModel, function_map: Dict[str, Callable]):
        assert (
            pydantic_available
        ), "Please install pydantic [pip3 install pydantic] to be able to use CommandInterpreter"
        self.validation_model = validation_model
        self.function_map = function_map
        self.context = {}

    def execute_action(self, action: Action):
        func = self.function_map.get(action.item)
        if not func:
            raise ValueError(f"Function {action.item} not found in function map.")

        # Execute function safely
        result = safe_execute(func, *action.args, **action.kwargs)
        self.context[action.result_var] = result

        # Handle post-processing if specified
        if action.post_processor and result["success"]:
            post_processor = eval(action.post_processor, {}, self.context)
            self.context[action.result_var] = {
                "success": True,
                "result": post_processor,
            }

    def execute(self, json_input: Union[dict, str, bytes]):
        try:
            if isinstance(json_input, (str, bytes)):
                json_input = json.loads(json_input)
            parsed_data = self.validation_model(**json_input)
        except Exception as e:
            return {"success": False, "error": str(e)}

        # Execute actions
        for action in parsed_data.actions:
            self.execute_action(action)

        # Handle return configurations
        if parsed_data.returns:
            for key, return_config in parsed_data.returns.items():
                return_fn = (
                    eval(return_config.fn, {}, self.context)
                    if return_config.fn
                    else None
                )
                args = [self.context[arg]["result"] for arg in return_config.args]
                self.context[return_config.result_var] = {
                    "success": True,
                    "result": return_fn(*args) if return_fn else args,
                }

        return self.context



import warnings
from pathlib import Path
from typing import (
    Any,
    Literal,
    Union,
    TypeGuard,
    TypeVar,
    Optional,
    Sequence,
    TypeAlias,
    Type,
)


try:
    from torch import Tensor  # type: ignore

    _has_torch = True
except ModuleNotFoundError:
    _has_torch = False

try:
    from numpy import ndarray  # type: ignore

    _has_numpy = True
except ModuleNotFoundError:
    _has_numpy = False


T = TypeVar("T")


def is_numpy(entry: Any, verbose_warn: bool = False) -> TypeGuard[T]:
    if not _has_numpy:
        if verbose_warn:
            warnings.warn(
                "You dont have numpy installed, the response will be always False"
            )
        return False
    return isinstance(entry, ndarray)


def is_tensor(entry: Any, verbose_warn: bool = False) -> TypeGuard[T]:
    if not _has_torch:
        if verbose_warn:
            warnings.warn(
                "You dont have torch installed, the response will be always False"
            )
        return False
    return isinstance(entry, Tensor)


def is_special_array(entry: Any) -> TypeGuard[T]:
    if not any([_has_numpy, _has_torch]):
        return False
    return is_numpy(entry, False) or is_tensor(entry, False)


def is_float(entry: Any, check_string: bool = False) -> TypeGuard[float]:
    """
    Checks if the entry provided is a valid float. It can check if a string can be converted to float if check_string is True.

    Args:
        entry (Any): The value to be checked.
        check_string (bool, optional): If True, it will check if the value can be converted to float.

    Returns:
        bool: If True, means that is a valid float otherwise false.
    """
    if isinstance(entry, float):
        return True
    if not check_string or not is_string(entry, False, True):
        return False
    try:
        entry = str(entry)
        float(entry)
        return True
    except:
        return False


def is_int(entry: Any, check_string: bool = False) -> TypeGuard[int]:
    """Checks if a number is a valid integer.

    Args:
        entry (Any): The item to be check if is a valid integer
        check_string (bool, optional): To check strings to see if its a possible number. Defaults to False.

    Returns:
        bool: True if its a True integer otherwise False.
    """
    if isinstance(entry, int):
        return True
    if not check_string or not is_string(entry, False, True):
        return False
    return str(entry).isdigit()


def is_dict(entry: Any, allow_empty: bool = False) -> TypeGuard[dict]:
    """Check if the provided entry is a valid dictionary and if it has content or not (if allow_empty is False).

    Args:
        entry (Any): The value to be checked if its True.
        allow_empty (bool, optional): If True it allow empty dictionaries to be evaluated, otherwise it requires it to be a dictionary and have at least some content there. Defaults to False.

    Returns:
        bool: True if valid dictionary and (if allow_empty is False) if it has some content in it.
    """
    return isinstance(entry, dict) and (allow_empty or bool(entry))


def is_number(entry: Any, check_string: bool = False) -> TypeGuard[Union[int, float]]:
    """Check if the entry is a number (being either a int or float). It also check if in case its a string (and check_string is True) if its a valid number if converted.

    Args:
        entry (Any): The entry to be checked.
        check_string (bool, optional): If True will check if entry can be converted to float, case the entry is a string. Defaults to False.

    Returns:
        bool: True if the entry is either a float or a integer, otherwise False.

    Examples:
        if for example 'entry' is a string and check_string is set to True:
        ```python
        a = "1"
        b = "text"
        c = "1.56"
        print(is_number(a, check_string=True)) # True
        print(is_number(a, check_string=False)) # False
        print(is_number(b, check_string=True)) # False
        print(is_number(b, check_string=False)) # False
        print(is_number(c, check_string=True)) # True
        print(is_number(c, check_string=False)) # False
        float(a) # ok
        float(b) # ValueError
        float(c) # ok
        ```

    """
    return is_int(entry, check_string) or is_float(entry, check_string)


def is_string(
    entry: Any,
    strip_string: bool = True,
    allow_empty: bool = False,
    check_path: bool = True,
) -> TypeGuard[str]:
    """Check if an entry is a string, bytes or a Path object.

    if ``strip_string`` is set to True, then the entry will be striped will be stripped before the final checking (Its useless when allow empty is False).
    """

    if not isinstance(entry, (str, bytes)):
        if not check_path or not isinstance(entry, Path):
            return False
        check_path = str(check_path)

    if not allow_empty and strip_string:
        entry = entry.strip()
    return allow_empty or bool(entry)


def is_list(entry: Any, allow_empty: bool = False) -> TypeGuard[list]:
    """Check if the provided entry is a list and if its not empty.

    Args:
        entry (Any): _description_
        allow_empty (bool, optional): If true, will return True if the item is a list, regardless if its empty or not. Defaults to False.

    Returns:
        bool: True if the entry is a list and if either the allow_empty is true or the list is not empty.
    """
    return isinstance(entry, list) and (allow_empty or bool(entry))


def is_tuple(entry: Any, allow_empty: bool = False) -> TypeGuard[tuple]:
    """Check if the provided entry is a valid dictionary and if it has content or not (if allow_empty is False).

    Args:
        entry (Any): The value to be checked if its True.
        allow_empty (bool, optional): If True it allow empty dictionaries to be evaluated, otherwise it requires it to be a dictionary and have at least some content there. Defaults to False.

    Returns:
        bool: True if valid dictionary and (if allow_empty is False) if it has some content in it.
    """
    return isinstance(entry, tuple) and (allow_empty or bool(entry))


def is_array(
    entry: Any, allow_empty: bool = False, **kwargs
) -> TypeGuard[Union[list, tuple, set]]:
    """Checks if the entry is either a list, tuple or set. It can also check for dictionaries if ``check_dict`` is True. Checks if its empty if allow_empty is False.

    Args:
        entry (Any): Value to be analyzed.
        allow_empty (bool, optional): If True will allow empty arrays to be returned as True. Defaults to False.
    Returns:
        bool: If True the value is a valid (non-empty if allow_empty is False else it returns true just for being a list or tuple).
    """
    if isinstance(entry, (list, tuple, set)):
        return allow_empty or bool(entry)
    return False


def is_bool(entry: Any, allow_number: bool = False) -> TypeGuard[bool]:
    if not allow_number or isinstance(entry, bool):
        return isinstance(entry, bool)
    return is_int(entry) and entry in [0, 1]


def default(entry: Optional[T], default: T) -> T:
    """
    entry if its not None or default.

    Useful to allow a different approach than 'or' operator in strings, for example:

    Consider that the arguments as:
    ```py
    arg1 = 0
    arg2 = 3
    ```
    If using or operator directly the following would happen:

    ```python
    results = arg1 or arg2
    # results = arg2 (3)
    ```
    It checks for Falsely data in the first item, but sometimes that value would be valid even if falsely like: `0`, `""`, `[]`, `{}`, `()` and `False`.

    So, it was made to check if the first value is None or non-None if None it uses the arg2, otherwise it returns the arg1 even if falsely.

    example:
    ```
    from gr1336_toolbox import default

    results_default = default(0, 3)
    # results_default = 0
    results_on_or = 0 or 3
    # results_on_or = 3
    results_default_2 = default(None, 0)
    # results_default_2 = 0
    ```

    """
    return entry if exists(entry) else default


def exists(val: Optional[T]) -> TypeGuard[T]:
    return val is not None


__all__ = [
    "is_int",
    "is_string",
    "is_dict",
    "is_float",
    "is_number",
    "is_tuple",
    "is_list",
    "is_bool",
    "is_array",
    "exists",
    "default",
    "is_numpy",
    "is_tensor",
    "is_special_array",
]



import os
import json
import yaml
import shutil
import traceback
from pathlib import Path
from .text import current_time
from .types import is_string, is_array, is_list
from typing import Any, Literal, Optional, Union, Sequence, List, Tuple, Dict
from .misc import flatten_list, filter_list


def scan_dir(
    path: Union[str, Path],
    pattern: Optional[str] = None,
    expected_dir: Literal["file", "path", "any"] = "any",
) -> list[Path]:
    path = Path(path)
    if expected_dir == "file":
        fn_checker = lambda path: Path.is_file(path)
    else:
        fn_checker = lambda path: Path.is_dir(path)
    if not pattern:
        pattern = "*"
    return [x for x in Path(path).glob(pattern) if fn_checker(x)]


def _validate_path_messages(
    entry: Union[str, Path],
    expected_dir: Literal["file", "path", "any"],
    error_type: Literal["unexpected_path", "path_not_exist", "invalid_type"],
):
    if error_type == "invalid_type":
        return f"Invalid type provided, received: '{type(entry)}' while expected 'str', 'bytes' or 'Path'."

    entry = str(entry).replace("\\", "/")
    if error_type == "path_not_exist":
        return 'The provided Path "{}" does not exist!'.format(entry)

    match expected_dir: # unexpected_path
        case "any":
            return 'The provided Path "{}" is not a valid directory or file.'.format(
                entry
            )
        case "file":
            return 'The provided Path "{}" is not a valid directory. It is a file not a directory!'.format(
                entry
            )
        case "path":
            return 'The provided Path "{}" is not a valid file. It is a directory and not a file!'.format(
                entry
            )


def validate_path(
    entry: Union[str, Path],
    expected_dir: Literal["file", "path", "any"] = "any",
    *,
    raise_exception: bool = False,
) -> bool:
    """Checks if `entry` is a valid existent path."""
    if not isinstance(entry, (str, bytes, Path)):
        if raise_exception:
            raise ValueError(_validate_path_messages(entry, expected_dir, "invalid_type"))
        return False
    if not Path(entry).exists():
        if raise_exception:
            raise ValueError(_validate_path_messages(entry, expected_dir, "path_not_exist"))
        return False
    entry = Path(entry)
    if expected_dir == "any":
        if raise_exception:
            raise ValueError(_validate_path_messages(entry, expected_dir, "unexpected_path"))
        return True
    else:
        is_file_type = entry.is_file()
        if expected_dir == "file":
            if not is_file_type and raise_exception:
                raise ValueError(_validate_path_messages(entry, expected_dir, "unexpected_path"))
            return is_file_type
        if is_file_type and raise_exception:
            raise ValueError(_validate_path_messages(entry, expected_dir, "unexpected_path"))
        return not is_file_type


def get_folders(
    path: Union[str, Path],
    pattern: str = "*",
    *args,
    **kwargs,
) -> list[Path] | list:
    if is_list(path):
        results = []
        paths = [x for x in path if validate_path(x, expected_dir="path")]
        if not paths:
            return []
        [
            results.extend(scan_dir(x, pattern=pattern, expected_dir="path"))
            for x in paths
        ]
        return list(sorted(results))
    if not validate_path(path, "path"):
        return []
    return scan_dir(path, pattern=pattern, expected_dir="path")


def _get_files_ext_set(extension: str):
    if extension.startswith("*."):
        return extension
    if extension.startswith("."):
        return "*" + extension
    return "*." + extension


def get_files(
    path: Union[List[Union[str, Path]], str, Path],
    extensions: Optional[Union[str, List[str], Tuple[str]]] = None,
    *args,
    **kwargs,
) -> list[Path] | list:
    results = []
    if is_list(path):
        paths = [Path(x) for x in path if validate_path(x, expected_dir="path")]
        if not paths:
            return results
        [results.extend(get_files(_path, extensions=extensions)) for _path in paths]
        return list(sorted(results))
    else:
        if not validate_path(path, expected_dir="path"):
            return results
        if is_array(extensions):
            [
                results.extend(
                    scan_dir(
                        path,
                        pattern=_get_files_ext_set(extension),
                        expected_dir="file",
                    )
                )
                for extension in extensions
            ]
        elif is_string(extensions):
            results.extend(
                scan_dir(
                    path,
                    pattern=_get_files_ext_set(extensions),
                    expected_dir="file",
                )
            )
        return list(sorted(results))


def path_to_string(path: Path):
    assert isinstance(path, (Path, str, bytes)), "Invalid Path format"
    return str(Path(path)).replace("\\", "/")


def mkdir(
    *paths: Union[Path, str],
):
    Path(*[x for x in paths if isinstance(x, (bytes, str, Path))]).mkdir(
        parents=True, exist_ok=True
    )


def setup_path(
    *paths: Union[str, Path],
    mkdir_path: bool = False,
    return_original: bool = False,
    **kwargs,
) -> Union[str, Path]:
    """
    The function `setup_path` takes in multiple paths as arguments, creates directories if specified,
    and returns the path as a string with forward slashes.

    Args:
      paths (str | Path): The paths that are to be managed.
      mkdir_path (bool): Create the directory specified in the path if it does not already exist. If `mkdir_path` is set to `True` and the directory does not exist, the function will create the directory. Defaults to False
      return_original (bool): Determines whether the function should return the type of the path (as a `Path` object) instead of the path as a string. If `return_original` is set to `True`, the function. Defaults to False

    Returns:
      The function `setup_path` returns a string representation of the path with backslashes replaced by forward slashes or as a `Path` object if `return_original` is set to True.
    """
    path = Path(*[x for x in paths if isinstance(x, (bytes, str, Path))])
    if mkdir_path and not path.exists():
        mkdir(m_path=path.parent if "." in path.name else path)
    if return_original:
        return path
    return str(path).replace("\\", "/")


def load_json(
    path: Union[str, Path],
    default_value: Optional[Any] = None,
    encoding: Optional[
        Union[str, Literal["utf-8", "ascii", "unicode-escape", "latin-1"]]
    ] = None,
    errors: Union[str, Literal["strict", "ignore"]] = "strict",
    *args,
    **kwargs,
) -> list | dict | None:
    """
    Load JSON/JSONL data from a file.

    Args:
        path (Union[str, Path]): The path to the JSON file.

    Returns:
        Union[list, dict, None]: The loaded JSON data as a list, dictionary, or None if any error occurs.
    """

    if not validate_path(path, expected_dir="file"):
        if default_value is None:
            raise FileNotFoundError(f"Invalid path '{path}'")
        return default_value
    path = Path(path)
    file = path.read_text(encoding=encoding, errors=errors)
    if path.name.endswith(".jsonl"):
        results = []
        for line in file.splitlines():
            try:
                results.append(json.loads(line))
            except Exception as e:
                pass
        return results
    try:
        return json.loads(file)
    except:
        return default_value


def save_json(
    path: Union[str, Path],
    content: Union[list, dict, tuple, map, str, bytes],
    indent: int = 4,
    *,
    encoding: Optional[
        Union[str, Literal["utf-8", "ascii", "unicode-escape", "latin-1"]]
    ] = None,
    errors: Union[str, Literal["strict", "ignore"]] = "strict",
    **kwargs,
) -> None:
    """
    Save JSON data to a file.

    Args:
        path (Union[str, Path]): The path to save the JSON file.
        content (Union[list, dict]): The content to be saved as JSON.
        encoding (str, optional): The encoding of the file. Defaults to "utf-8".
        indent (int, optional): The indentation level in the saved JSON file. Defaults to 4.
    """

    if not is_string(path):
        path = current_time() + ".json"
    path = Path(path)
    if not path.name.endswith((".json", ".jsonl")):
        path = Path(path.parent, f"{path.name}.json")
    mkdir(Path(path).parent)
    if path.name.endswith(".jsonl"):
        if not isinstance(content, (str, bytes)):
            content = json.dumps(content)
        if path.exists():
            older_content = path.read_text(encoding=encoding, errors=errors)
            content = older_content.rstrip() + "\n" + content
    else:
        content = json.dumps(content, indent=indent)
    path.write_text(content, encoding=encoding, errors=errors)


def load_text(
    path: Union[Path, str],
    *,
    encoding: Optional[
        Union[str, Literal["utf-8", "ascii", "unicode-escape", "latin-1"]]
    ] = None,
    errors: Union[str, Literal["strict", "ignore"]] = "strict",
    default_value: Optional[Any] = None,
    **kwargs,
) -> str:
    if not validate_path(path, expected_dir="file"):
        if default_value is None:
            raise FileNotFoundError(f"Invalid path '{path}'")
        return default_value
    return Path(path).read_text(encoding, errors=errors)


def save_text(
    path: Union[Path, str],
    content: str,
    *,
    encoding: Optional[
        Union[str, Literal["utf-8", "ascii", "unicode-escape", "latin-1"]]
    ] = None,
    errors: Union[str, Literal["strict", "ignore"]] = "strict",
    **kwargs,
) -> None:
    """Save a text file to the provided path."""
    path = Path(path)
    mkdir(Path(path).parent)
    path.write_text(content, encoding=encoding, errors=errors)


def load_yaml(
    path: Union[Path, str],
    *,
    default_value: Any | None = None,
    safe_loader: bool = False,
    **kwargs,
) -> Optional[Union[List[Any], Dict[str, Any]]]:
    """
    Loads YAML content from a file.

    Args:
        path (Union[Path, str]): The path to the file.
        default_value (Any | None): If something goes wrong, this value will be returned instead.
        safe_loader (bool): If True, it will use the safe_load instead.

    Returns:
        Optional[Union[List[Any], Dict[str, Any]]]: The loaded YAML data.
    """
    if not validate_path(path, expected_dir="file"):
        if default_value is None:
            raise FileNotFoundError(f"Invalid path '{path}'")
        return default_value
    loader = yaml.safe_load if safe_loader else yaml.unsafe_load
    try:
        return loader(Path(path).read_bytes())
    except Exception as e:
        print(f"YAML load error: {e}")
        print("----------------------")
        raise e


def save_yaml(
    path: Union[Path, str],
    content: Union[List[Any], Tuple[Any, Any], Dict[Any, Any]],
    *,
    encoding: Optional[
        Union[str, Literal["utf-8", "ascii", "unicode-escape", "latin-1"]]
    ] = None,
    errors: Union[str, Literal["strict", "ignore"]] = "strict",
    safe_dump: bool = False,
    **kwargs,
) -> None:
    """Saves a YAML file to the provided path.

    Args:
        path (Union[Path, str]): The path where the file will be saved.
        content (Union[List[Any], Tuple[Any, Any], Dict[Any, Any]]): The data that will be written into the file.
        encoding (str, optional): The encoding of the output. Default is 'utf-8'. Defaults to "utf-8".
        safe_dump (bool, optional): If True, it uses the safe_dump method instead. Defaults to False.
    """
    mkdir(Path(path).parent)
    save_func = yaml.safe_dump if safe_dump else yaml.dump
    try:
        with open(path, "w", encoding=encoding, errors=errors) as file:
            save_func(data=content, stream=file, encoding=encoding)
    except Exception as e:
        print(f"An exception occurred while saving {path}. Exception: {e}")
        traceback.print_exc()


def move(
    source: Union[str, Path],
    destination: Union[str, Path],
    *args,
    **kwargs,
):
    """
    Moves a file or directory from one location to another.

    Args:
        source_path (Union[str, Path]): The path of the file/directory to be moved.
        destination_path (Union[str, Path]): The new location for the file/directory.

    Raises:
        AssertionError: If the source path does not exist or is invalid
    """
    assert str(source).strip() and Path(source).exists(), "Source path does not exists!"
    source = Path(source)
    assert validate_path(source), "Source path does not exists!"
    mkdir(destination)
    shutil.move(str(source), str(destination))


def delete(
    files: Union[str, Path, Sequence[Union[str, Path]]],
    verbose: bool = False,
    *args,
    **kwargs,
):
    if is_string(files) and Path(files).exists():
        files = Path(files)
        if files.is_dir():
            shutil.rmtree(str(files))
        else:
            os.rmdir(str(files))
        if verbose:
            files = path_to_string(files)
            print(f"'{files}' deleted")
    elif is_array(files):
        [delete(path) for path in filter_list(flatten_list(files), (str, Path))]


__all__ = [
    "get_folders",
    "get_files",
    "validate_path",
    "mkdir",
    "load_json",
    "save_json",
    "load_text",
    "save_text",
    "load_yaml",
    "save_yaml",
    "move",
    "delete",
    "setup_path",
    "scan_dir",
]
